/**
  "Reversible" computer simulator.  The idea is assignments destroy information,
  burning energy, so they're not normally allowed.
  
  Arithmetic is OK, pass by reference is OK, output is OK (burn the bits)
  
  See the philosophy of reversible computing here:
  http://www.cise.ufl.edu/research/revcomp/faq.html
  
  Modern survey here:
  http://www.informatik.uni-bremen.de/agra/doc/konf/2012_vdat_reversible_circuits_accompl_chall.pdf
  
  
*/
#ifndef __OSL_rev_INT_H
#define __OSL_rev_INT_H
#include <iostream>
#include <stdlib.h> /* for rand() */
#include <string.h> /* for memset() */

class rev_int {
private:
	int v; // the underlying value is a normal int, but you can't read it directly.
	int orig_v; // for reversability checking, v should return to this value
	
	// note there is NOT a copy constructor or assignment operator--these aren't allowed!
	rev_int(const rev_int &src); // deleted
	void operator=(const rev_int &src); // deleted
public:
	rev_int() { v=rand(); orig_v=v; } // <- unitialized values are filled with entropy
	rev_int(int v_) :v(v_) { orig_v=v; } // <- allows us to create program constants and such
	
	// Destructor checks to verify reversibility
	~rev_int() {
		if (v!=orig_v) {
			std::cout<<"Non-reversible computation: got "<<v<<" instead of "<<orig_v<<" for int "<<&v<<"\n";
		}
	}
	
	// You do get swap: on adiabatic hardware this would be 
	//   implemented using crossing wires, not assignments, so it is power-efficient.
	friend void swap(rev_int &a,rev_int &b) {
		int temp=a.v; a.v=b.v; b.v=temp;
	}
	
	// Self-arithmetic works normally.  Add and subtract are reversible.
	rev_int &operator+=(const rev_int &a) { v+=a.v; return *this; }
	rev_int &operator-=(const rev_int &a) { v-=a.v; return *this; }
	
	// Multiplication is perfectly reversible (ignoring overflow);
	//  int division is technically only reversible when remainder is zero.
	rev_int &operator*=(const rev_int &a) { v*=a.v; return *this; }
	rev_int &operator/=(const rev_int &a) { v/=a.v; return *this; }
	
	rev_int &operator^=(const rev_int &a) { v^=a.v; return *this; } // xor
	
	// Boolean comparisons work normally (arguably should use a reversible predicate for the return value)
	friend bool operator< (const rev_int &a,const rev_int &b) { return a.v<b.v; }
	friend bool operator<=(const rev_int &a,const rev_int &b) { return a.v<=b.v; }
	friend bool operator> (const rev_int &a,const rev_int &b) { return a.v>b.v; }
	friend bool operator>=(const rev_int &a,const rev_int &b) { return a.v>=b.v; }
	friend bool operator==(const rev_int &a,const rev_int &b) { return a.v==b.v; }
	friend bool operator!=(const rev_int &a,const rev_int &b) { return a.v!=b.v; }
	
	// Printing stuff works.  You'd pay some bit erasure cost here, but that's cheap compared to the I/O.
	//   Saving your outputs instead of reversing them is the "Bennett trick" after [Bennett, 1973]
	friend std::ostream & operator<<(std::ostream &o,const rev_int &a) { return o<<a.v; }
	
	friend class rev_int_array;
};


// An array of reversible integers
class rev_int_array {
	rev_int *arr;
	int len;
	
	// deleted copy constructors
	rev_int_array(const rev_int_array &);
	void operator=(const rev_int_array &);
public:
// Allocate a brick of constant zeros.  Compared to new rev_int[],
//  This is an expensive operation, because it overwrites what was there before with zeros.
	rev_int_array(const rev_int &n) {
		len=n.v;
		arr=new rev_int[len]; // full of entropy
		memset(arr,0,len*sizeof(rev_int)); // now grounded to zero
	}
	~rev_int_array() {
		for (int i=0;i<len;i++) {
			if (arr[i].v!=0) {
				std::cout<<"non-reversible: arr["<<i<<"]="<<arr[i]<<"\n";
			}
		}
		delete[] arr;
		arr=0;
	}
	
	rev_int &operator[](const rev_int &index) { 
		if (index.v>=len || index.v<0) std::cout<<"rev_int_array out of bounds at "<<index.v<<" (should be under "<<len<<")\n";
		return arr[index.v];
	}
	
	// I'm too lazy to implement array swap, but it's legal.
};


#endif

